#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""FAL command for Clone."""

from __future__ import with_statement

import os
import random
from datetime import time
from ionas import request
from ionas import nascmd
from iobase.exception import (NotFoundError,)
from libfal import (falcmd, util)
import pal_info
import fal_config as config
from fal_info import (ReqTypes, Results)
import product_conf
from iobase.xmldata import (XmlNode, XmlConfNode, escape, XmlData)
from iobase.netut import is_inet4_addr, is_inet6_addr
from iobase.cmdexec import (CmdExec, NULL)
from iobase.fileut import (FLock,)
from pkgaddon.fal_hclone_config import (CLONE_WATCHDOG_SERVICE,
                                        CLONE_SYSTEM_BACKUP_NAME)
from pkgaddon.hclone_config import (BACKUP_NEW_XML_FW_VER,)


ENABLE = 'enable'
MASTER = 'master'
SLAVE = 'slave'
SYSTEM = 'system'
HOST = 'host'
PASSWD = 'passwd'
WATCH = 'watch'
DEADTIME = 'deadtime'
TIMES = 'times'
SCHEDULE = 'schedule'
TIME = 'time'
SIGNATURE = 'signature'
EXISTS = "exists"

SIGNATURE_XML_TMPL = "<signature>%s</signature>"

NETIF_PATH = "network.v2/netiflist/netif"
INET4_PROTOCOL_PATH = NETIF_PATH + "/protocollist/protocol[@family='inet4']"
INET6_PROTOCOL_PATH = NETIF_PATH + "/protocollist/protocol[@family='inet6']"
INET4_ADDRS_PATH = INET4_PROTOCOL_PATH + "/addrlist/addr"
INET6_ADDRS_PATH = INET6_PROTOCOL_PATH + "/addrlist/addr"
HOSTNAME_PATH = "sysinfo/hostdesc/name"

SLAVE_WATCH_PATH = "slave/watch"

# Add the service entry for the clone package
util.config.FUNC_TO_SERVICE_MAP[CLONE_WATCHDOG_SERVICE] = \
                                        CLONE_WATCHDOG_SERVICE

IMMEDIATE_OPT = "immediate"
TRUE_TEXT = "true"
UNDELETABLE_ATTR = "undeletable"

SHARE_USER_INFO_TAG_LIST = ("uid", "uname", "gid")

PING_OPTS = ["-c", "1"]

DROPBOX_PATH = "shares/sharelist/share/shareservice/dropbox"


class FalClone(falcmd.FalCmd):
    """FAL Clone command implementation class."""

    CLONE_CONF_FILE = os.path.join(config.XML_CONF_DIR, 'config-clone.xml')
    FEATURES_CONF_FILE = os.path.join(
        config.XML_CONF_DIR, 'config-features.xml')

    USB_NO_STR = sorted(config.USB_PORT_NO_LIST, reverse=True)[0]

    def __load_config_file(self, config_file):
        """Load "config_file" and return the response's entry."""

        return self.send_entry_with_file(
                    request.Entry(ReqTypes.LOAD),
                    pal_info.PAL_CONFFILE_PATH, config_file)

    def __save_config_file(self, config_file, config_xml):
        """Save "xml_data" to "config_file" and return the response's entry."""

        return self.send_entry_with_file(
                    request.Entry(ReqTypes.SAVE, config_xml),
                    pal_info.PAL_CONFFILE_PATH, config_file)

    def __logging(self, constname, arglist={}):
        try:
            from subprocess import Popen, PIPE
            proc = Popen(
                    ['/usr/local/bin/eventlog', '-c', constname] + ["%s=%s" % (k, v) for k, v in arglist.iteritems()],
                    stdout=PIPE,
                    stderr=PIPE)
            (stdout, stderr) = proc.communicate()
        except:
            pass

    def __alive(self, target):
        try:
            PAL_CLONE_PATH = '/mnt/data/bin/pal_clone.py'
            entry = self.send_entry(
                    request.Entry(
                            'alive',
                            """<clone>
                                <target>%s</target>
                            </clone>""" % escape(target)),
                    PAL_CLONE_PATH)
            return entry.get_result() == '0'
        except:
            return False

    def __usb_is_share(self):
        try:
            def __fal_ports_load():
                return self.send_entry_with_file(
                        request.Entry(ReqTypes.LOAD),
                        pal_info.PAL_CONFFILE_PATH,
                        config.PORTS_CONF_FILE)

            entry = __fal_ports_load()
            return entry.get().get_node(
                    'usblist/usb[@port="%s"]/modelist/mode[text()="share"]' %
                        (self.USB_NO_STR,)).is_enabled()
        except:
            return False

    def __usb_is_mounted(self):
        try:
            return os.path.exists('/dev/usbd%s' % (self.USB_NO_STR,))
        except:
            return False

    def __usb_is_ext4(self):
        try:
            entry = self.send_entry(
                    request.Entry(
                            ReqTypes.QUERY,
                            """<volctrl>
                                <volume>
                                    <name type="disk">usbd%s</name>
                                </volume>
                            </volctrl>""" % (self.USB_NO_STR,)),
                    pal_info.PAL_VOLCTRL_PATH)
            return entry.get().get_value('volume/filesystem') in ['ext4']
        except:
            return False

    def __backupdata_is_exists(self):
        try:
            return os.path.isfile(self.LANDISKINFO_DATA_FILE_PATH)
        except:
            return False

    def __usb_is_backup_ok(self):
        try:
            return self.__usb_is_share() \
                    and self.__usb_is_mounted() \
                    and self.__usb_is_ext4()
        except:
            return False

    def __usb_is_restore_ok(self):
        try:
            return self.__backupdata_is_exists()
        except:
            return False

    def __is_available_replication(self):
        """Check replication feature availability."""

        FUNCTION_XPATH = "functionlist/function"
        VERSION_ATTR = "version"
        REPLICATION_FEATURE = "replication"
        OLDEST_VERSION = 1

        try:
            features = self.send_entry_with_file(
                request.Entry(ReqTypes.LOAD),
                pal_info.PAL_CONFFILE_PATH,
                self.FEATURES_CONF_FILE).get()
            for node in features.get_node_list(FUNCTION_XPATH):
                version = float(node.get_attr_value(VERSION_ATTR, "0"))
                if (node.get_text() == REPLICATION_FEATURE and
                        version >= OLDEST_VERSION):

                    return True
        except:
            pass

        return False

    def __load_config(self):
        e_conf = self.send_entry_with_file(
            request.Entry(ReqTypes.LOAD),
            pal_info.PAL_CONFFILE_PATH,
            self.CLONE_CONF_FILE)

        n_clone = e_conf.get()

        def _is_enabled(node, xpath, default=False):
            try:
                return node.get_node(xpath).is_enabled()
            except:
                return default

        data = dict()
        if n_clone.get_node_list('./master'):
            data[MASTER] = {
                ENABLE: _is_enabled(n_clone, './master'),
                SLAVE: {
                    HOST: n_clone.get_value('./master/slave/host'),
                    PASSWD: n_clone.get_value('./master/slave/passwd'),
                }
            }
        else:
            data[MASTER] = None
        if n_clone.get_node_list('./slave'):
            data[SLAVE] = {
                WATCH: {
                    ENABLE: _is_enabled(n_clone, './slave/watch'),
                    MASTER: {
                        HOST: n_clone.get_value('./slave/watch/master/host'),
                    }
                }
            }
        else:
            data[SLAVE] = None
        if n_clone.get_node_list('./system'):
            data[SYSTEM] = {
                ENABLE: _is_enabled(n_clone, './system'),
                TIMES: int(n_clone.get_value('./system/times', '0')),
                SCHEDULE: {
                    ENABLE: _is_enabled(n_clone, './system/schedule'),
                    TIME: time(
                        int(n_clone.get_value('./system/schedule/time/hour', '0')),
                        int(n_clone.get_value('./system/schedule/time/minute', '0'))),
                }
            }
        else:
            data[SYSTEM] = None

        return data

    def __delete_userinfo_for_share(self, share):
        """Delete tags(uid, uname and gid) in "share"."""

        for tag_name in SHARE_USER_INFO_TAG_LIST:
            share.delete_children(tag_name)

    # Handlers ###

    def __attach_switch_ok(self, entry):
        try:
            # FIXME file path
            #ok = os.path.isfile('/mnt/data/conf/clone/landiskinfo.xml')
            ok = os.path.isfile(self.LANDISKINFO_XML_FILE_PATH)

            clone_node = entry.get()
            if not clone_node.get_node_list('./slave'):
                clone_node.append_child(XmlNode('<slave />'))
            clone_node.get_node('./slave').set_bool_attr('switch', true=ok)
        except:
            pass

    def __attach_restore_ok(self, entry):
        try:
            ok = self.__usb_is_restore_ok()

            clone_node = entry.get()
            if not clone_node.get_node_list('./system'):
                clone_node.append_child(XmlNode('<system />'))
            clone_node.get_node('./system').set_bool_attr('restore', true=ok)
        except:
            pass

    def __attach_saveconfig_restore(self, entry):
        """
        <saveconfig>
            <config for="restore">
                <hostname>(HOSTNAME)</hostname>
                <timestamp>(TIMESTAMP)</timestamp>
            </config>
        </saveconfig>
        """
        try:
            landiskinfo_node = self.__load_backup_config_xml()
            hostname = landiskinfo_node.get_value('network/host/name', '')
            timestamp = landiskinfo_node.get_attr_value('update')
            ipaddr = {
                'lan1': {
                    'ipv4': landiskinfo_node.get_value(
                            'network/tcpip[@family="inet"]/lan/ipsetting/ip',
                            ''),
                    'ipv6': landiskinfo_node.get_value(
                            'network/tcpip[@family="inet6"]/lan/ipsetting/ip',
                            ''),
                },
                'lan2': {
                    'ipv4': landiskinfo_node.get_value(
                            'network/tcpip[@family="inet"]/ext/ipsetting/ip',
                            ''),
                    'ipv6': landiskinfo_node.get_value(
                            'network/tcpip[@family="inet6"]/ext/ipsetting/ip',
                            ''),
                },
            }

            clone_node = entry.get()
            if not clone_node.get_node_list('./saveconfig'):
                clone_node.append_child(XmlNode('<saveconfig />'))
            saveconfig_node = clone_node.get_node('saveconfig')
            if not saveconfig_node.get_node_list('./config[@for="restore"]'):
                saveconfig_node.append_child(XmlNode('<config for="restore" />'))
            config_node = saveconfig_node.get_node('./config[@for="restore"]')
            config_node.append_child(XmlNode("<hostname>%s</hostname>" % escape(hostname)))
            config_node.append_child(XmlNode("<timestamp>%s</timestamp>" % escape(timestamp)))
            config_node.append_child(XmlNode('<ipaddr lan="1" ipv="4">%s</ipaddr>' % escape(ipaddr['lan1']['ipv4'])))
            config_node.append_child(XmlNode('<ipaddr lan="1" ipv="6">%s</ipaddr>' % escape(ipaddr['lan1']['ipv6'])))
            config_node.append_child(XmlNode('<ipaddr lan="2" ipv="4">%s</ipaddr>' % escape(ipaddr['lan2']['ipv4'])))
            config_node.append_child(XmlNode('<ipaddr lan="2" ipv="6">%s</ipaddr>' % escape(ipaddr['lan2']['ipv6'])))
        except:
            pass

    def __attach_saveconfig_switch(self, entry):
        """
        <saveconfig>
            <config for="switch">
                <hostname>(HOSTNAME)</hostname>
                <timestamp>(TIMESTAMP)</timestamp>
            </config>
        </saveconfig>
        """
        try:
            landiskinfo_node = self.__load_save_config_xml()
            hostname = landiskinfo_node.get_value('network/host/name', '')
            timestamp = landiskinfo_node.get_attr_value('update')
            ipaddr = {
                'lan1': {
                    'ipv4': landiskinfo_node.get_value(
                            'network/tcpip[@family="inet"]/lan/ipsetting/ip',
                            ''),
                    'ipv6': landiskinfo_node.get_value(
                            'network/tcpip[@family="inet6"]/lan/ipsetting/ip',
                            ''),
                },
                'lan2': {
                    'ipv4': landiskinfo_node.get_value(
                            'network/tcpip[@family="inet"]/ext/ipsetting/ip',
                            ''),
                    'ipv6': landiskinfo_node.get_value(
                            'network/tcpip[@family="inet6"]/ext/ipsetting/ip',
                            ''),
                },
            }

            clone_node = entry.get()
            if not clone_node.get_node_list('./saveconfig'):
                clone_node.append_child(XmlNode('<saveconfig />'))
            saveconfig_node = clone_node.get_node('saveconfig')
            if not saveconfig_node.get_node_list('./config[@for="switch"]'):
                saveconfig_node.append_child(XmlNode('<config for="switch" />'))
            config_node = saveconfig_node.get_node('./config[@for="switch"]')
            config_node.append_child(XmlNode("<hostname>%s</hostname>" % escape(hostname)))
            config_node.append_child(XmlNode("<timestamp>%s</timestamp>" % escape(timestamp)))
            config_node.append_child(XmlNode('<ipaddr lan="1" ipv="4">%s</ipaddr>' % escape(ipaddr['lan1']['ipv4'])))
            config_node.append_child(XmlNode('<ipaddr lan="1" ipv="6">%s</ipaddr>' % escape(ipaddr['lan1']['ipv6'])))
            config_node.append_child(XmlNode('<ipaddr lan="2" ipv="4">%s</ipaddr>' % escape(ipaddr['lan2']['ipv4'])))
            config_node.append_child(XmlNode('<ipaddr lan="2" ipv="6">%s</ipaddr>' % escape(ipaddr['lan2']['ipv6'])))
        except:
            pass

    def __attach_generationlist(self, entry):
        try:
            backups_node = XmlNode(self.__get_backups_xml())
            job_node = backups_node.get_node('backup/job')
            job_node.get_node('destination').append_child(XmlNode('<path>%s</path>' % ('/mnt/%s' % self.BACKUP_JOB_DEST)))

            _entry = self.send_entry(
                    request.Entry(
                            ReqTypes.QUERY,
                            """<backups>
                                <option detail="true">
                                    <joblist>%s</joblist>
                                </option>
                            </backups>""" % str(job_node)),
                    pal_info.PAL_BACKUPS_PATH)

            entry.get().append_child(_entry.get().get_node('backup/joblist/job/id[text()="s1"]/../generationlist'))
        except:
            pass

    def __attach_saved_signature_info(self, entry):
        try:
            signature_node = entry.get().get_node(SIGNATURE)
            is_in_usb = False
            if self.__read_signature() == signature_node.get_text():
                is_in_usb = True

            signature_node.set_bool_attr(EXISTS, is_in_usb)
        except:
            # Ignore error
            pass

    def __query_handler(self, entry):
        """Handle query request.
        <entry type="query">
            <clone />
        </entry>
        """
        entry = entry.get_clone()

        _entry = self.send_entry_with_file(
                request.Entry(ReqTypes.LOAD),
                pal_info.PAL_CONFFILE_PATH,
                self.CLONE_CONF_FILE)

        self.__attach_switch_ok(_entry)
        self.__attach_restore_ok(_entry)
        self.__attach_saveconfig_restore(_entry)
        self.__attach_saveconfig_switch(_entry)
        self.__attach_generationlist(_entry)
        self.__attach_saved_signature_info(_entry)

        entry.set(_entry.get())

        entry.set_result(Results.OK)
        return entry

    def __enable_replication(self, shost, spasswd):
        # FIXME イベントハンドラ
        try:
            entry = self.send_entry_with_file(
                    request.Entry(ReqTypes.LOAD),
                    pal_info.PAL_CONFFILE_PATH,
                    config.SHARES_CONF_FILE)
            if entry.get_result() != Results.OK:
                return []
            shares_node = entry.get()
            changelist = []
            for share_node in shares_node.get_node_list('sharelist/share'):
                name = share_node.get_value('name', None)
                if name is None:
                    continue
                if share_node.is_set_to_true(UNDELETABLE_ATTR):
                    continue
                self.__delete_userinfo_for_share(share_node)
                share_node.delete_children('replication')
                replication_node = XmlConfNode('<replication />')
                replication_node.enable()
                for tagname, text in [
                        ('host', shost),
                        ('passwd', spasswd),
                        ('share', name),
                        ('username', 'admin')]:  # FIXME admin?
                    node = XmlNode("<%s />" % tagname)
                    node.set_text(text)
                    replication_node.append_child(node)
                share_node.append_child(replication_node)
                _shares_node = XmlNode('<shares />')
                _shares_node.append_child(share_node)
                from fal_info import FAL_SHARES_PATH
                self.send_entry(
                        request.Entry(ReqTypes.MODIFY, _shares_node),
                        FAL_SHARES_PATH)
                changelist.append(name)
            return changelist
        except:
            return []

    def __disable_replication(self):
        """Disable the replication setting in the share configuration file."""

        try:
            shares_config = self.__load_config_file(
                                        config.SHARES_CONF_FILE).get()
            for share_node in shares_config.get_node_list('sharelist/share'):
                name = share_node.get_value('name', None)
                if name is None:
                    continue
                if share_node.is_set_to_true(UNDELETABLE_ATTR):
                    continue

                try:
                    share_node.get_node('replication').enable(False)
                except:
                    continue

                self.__delete_userinfo_for_share(share_node)
                _shares_node = XmlNode('<shares />')
                _shares_node.append_child(share_node)
                from fal_info import FAL_SHARES_PATH
                self.send_entry(
                        request.Entry(ReqTypes.MODIFY, _shares_node),
                        FAL_SHARES_PATH)
        except:
            pass

        return

    def __apply_handler(self, entry):
        """Handle apply request.
        <entry type="apply">
            <clone>
                <master enable="(true|false)">
                    <slave>
                        <host>(HostName|IPAddress)</host>
                        <passwd>(Password)</passwd>
                    </slave>
                </master>
                <slave>
                    <watch enable="(true|false)">
                        <deadtime>(10~60)</deadtime>
                        <master>
                            <host>(HostName|IPAddress)</host>
                        </master>
                    </watch>
                </slave>
                <system>
                    <times>0</times>
                    <schedule enable="(true|false)">
                        (<day>(Sun|Mon|Tue|Wed|Thu|Fri|Sat)</day>){1,7}
                        <time>
                            <hour>(00-23)</hour>
                            <minute>(00-55)</minute>
                        </time>
                    </schedule>
                </system>
            </clone>
        </entry>

        response result
        0(Results.OK)            : OK
        11                       : USB is disconnect
        12                       : USB is not share
        13                       : USB is not ext4
        21                       : Master is dead
        4(Results.INTERNAL_ERROR): NG
        """
        entry = entry.get_clone()
        clone_node = entry.get()

        data = dict()
        if clone_node.get_node_list('./master'):
            data[MASTER] = {
                ENABLE: clone_node.get_node('./master').is_enabled(),
                SLAVE: {
                    HOST: clone_node.get_node(
                            './master/slave/host').get_text(),
                    PASSWD: clone_node.get_node(
                            './master/slave/passwd').get_text(),
                }
            }
        else:
            data[MASTER] = None
        if clone_node.get_node_list('./slave'):
            data[SLAVE] = {
                WATCH: {
                    ENABLE: clone_node.get_node('./slave/watch').is_enabled(),
                    DEADTIME: clone_node.get_node(
                            './slave/watch/deadtime').get_text(),
                    MASTER: {
                        HOST: clone_node.get_node(
                                './slave/watch/master/host').get_text(),
                    }
                }
            }
            if data[SLAVE][WATCH][ENABLE]:
                if not self.__alive(data[SLAVE][WATCH][MASTER][HOST]):
                    entry.set_result("21")
                    return entry

                util.enable_service(CLONE_WATCHDOG_SERVICE, True)
            else:
                util.enable_service(CLONE_WATCHDOG_SERVICE, False)
                util.servicectrl_stop([], [CLONE_WATCHDOG_SERVICE])
        else:
            data[SLAVE] = None
        if clone_node.get_node_list('./system'):
            data[SYSTEM] = {
                TIMES: int(clone_node.get_value('./system/times', '0')),
                SCHEDULE: {
                    ENABLE: clone_node.get_node(
                            './system/schedule').is_enabled(),
                    TIME: time(
                        int(clone_node.get_node(
                                './system/schedule/time/hour').get_text()),
                        int(clone_node.get_node(
                                './system/schedule/time/minute').get_text())),
                }
            }
            if data[SYSTEM][SCHEDULE][ENABLE]:
                if not self.__usb_is_mounted():
                    entry.set_result("11")
                    return entry
                if not self.__usb_is_share():
                    entry.set_result("12")
                    return entry
                if not self.__usb_is_ext4():
                    entry.set_result("13")
                    return entry
        else:
            data[SYSTEM] = None

        _clone_node = self.send_entry_with_file(
            request.Entry(ReqTypes.LOAD),
            pal_info.PAL_CONFFILE_PATH,
            self.CLONE_CONF_FILE).get()
        if data[MASTER]:
            _clone_node.delete_children('master')
            _clone_node.append_child(clone_node.get_node('./master').get_clone())
            if data[MASTER][ENABLE] and not self.__is_available_replication():
                entry.set_result(Results.NO_FEATURE_ERROR)
                return entry
        if data[SLAVE]:
            _clone_node.delete_children('slave')
            _clone_node.append_child(clone_node.get_node('./slave').get_clone())
        if data[SYSTEM]:
            _clone_node.delete_children('system')
            _clone_node.append_child(clone_node.get_node('./system').get_clone())

        if Results.OK == self.send_entry(
                entry,
                getattr(
                        pal_info,
                        'PAL_CLONE_PATH',
                        '/mnt/data/bin/pal_clone.py'),
                ).get_result():

            if data[MASTER] and data[MASTER][ENABLE]:
                shost = data[MASTER][SLAVE][HOST]
                spasswd = data[MASTER][SLAVE][PASSWD]
                _entry = request.Entry(ReqTypes.SEND, """<clone>
                            <slave>
                                <host>%(host)s</host>
                                <passwd>%(passwd)s</passwd>
                            </slave>
                        </clone>""" % {
                                'host': escape(shost),
                                'passwd': escape(spasswd),
                            })
                _entry.set_options({IMMEDIATE_OPT: TRUE_TEXT})
                if Results.OK != self.send_entry(_entry,
                        pal_info.PAL_CLONE_PATH).get_result():

                    entry.set_result(Results.INTERNAL_ERROR)
                    return entry

                self.__enable_replication(shost, spasswd)

            if data[MASTER] and not data[MASTER][ENABLE]:
                self.__disable_replication()

            if Results.OK == self.send_entry_with_file(
                    request.Entry(ReqTypes.SAVE, _clone_node),
                    pal_info.PAL_CONFFILE_PATH,
                    self.CLONE_CONF_FILE).get_result():
                entry.set_result(Results.OK)

                if data[SLAVE] and data[SLAVE][WATCH][ENABLE]:
                    util.servicectrl_start([], [CLONE_WATCHDOG_SERVICE])
            else:
                entry.set_result(Results.INTERNAL_ERROR)  # FIXME code
        else:
            entry.set_result(Results.INTERNAL_ERROR)  # FIXME code

        #entry.set_result(Results.OK)
        return entry

    # for MasterSlave ###
    LANDISKINFO_XML_FILE_PATH = '/mnt/data/conf/clone/landiskinfo.xml'

    def __load_save_config_xml(self):
        try:
            landiskinfo_node = None
            with open(self.LANDISKINFO_XML_FILE_PATH, 'r') as f:
                landiskinfo_node = XmlNode(f.read())
                landiskinfo_node.delete_nodes(
                    'raidctrl/raidlist',
                    'raid/name[text()="sataraid1"]/..')
            return landiskinfo_node
        except:
            return None

    def __save_handler(self, entry):
        """Handle save request.
           In Slave.
           Execute from setslaveinfo.
           Save "<landiskinfo ~~~ />" to /mnt/data/conf/clone/xxx.xml.
        <entry type="save">
            <landiskinfo>
                ~~~
            </landiskinfo>
        </entry>
        """
        entry = entry.get_clone()

        entry.set_result(Results.OK)
        return entry

    def __is_valid_master_config(self):
        """Check monitoring master config and master data."""

        try:
            config_master = self.__load_config()[SLAVE][WATCH][MASTER][HOST]
            if not config_master:
                return True
        except:
            return True

        try:
            landiskinfo_node = self.__load_save_config_xml()
            if is_inet4_addr(config_master):
                addr_node_list = landiskinfo_node.get_node_list(
                    INET4_ADDRS_PATH)
                for addr_node in addr_node_list:
                    tmp_addr = addr_node.get_text()
                    if tmp_addr == config_master:
                        return True
            elif is_inet6_addr(config_master):
                addr_node_list = landiskinfo_node.get_node_list(
                    INET6_ADDRS_PATH)
                for addr_node in addr_node_list:
                    tmp_addr = addr_node.get_text()
                    if tmp_addr == config_master:
                        return True
            else:
                hostname = landiskinfo_node.get_node(HOSTNAME_PATH).get_text()
                if hostname.lower() == config_master.lower():
                    return True
        except:
            pass

        return False

    def __try_update_service_for_switch(self):
        """Update condition and configuration for service."""

        try:
            clone_config = self.__load_config_file(self.CLONE_CONF_FILE).get()
            watch = clone_config.get_node(SLAVE_WATCH_PATH)
            if not watch.is_enabled():
                return
        except:
            return

        util.enable_service(CLONE_WATCHDOG_SERVICE, False)
        util.servicectrl_stop([], [CLONE_WATCHDOG_SERVICE])
        watch.enable(False)

        self.__save_config_file(self.CLONE_CONF_FILE, clone_config)

    def __ping_to_target(self, ping_cmd, target):
        """Execute "ping_cmd" to the "target"."""

        CmdExec([ping_cmd] + PING_OPTS + [target],
                    stdout=NULL, stderr=NULL).execute()

    def __is_hostname_or_ip_used(self):
        """Check if host name or IP address is used."""

        try:
            landiskinfo_data = XmlData(self.__load_save_config_xml())

            hostname = landiskinfo_data.get_node(HOSTNAME_PATH).get_text()
            inet4_addr_list = landiskinfo_data.get_text_list(INET4_ADDRS_PATH)
            inet6_addr_list = landiskinfo_data.get_text_list(INET6_ADDRS_PATH)

            for target in [hostname] + inet4_addr_list:
                if target:
                    try:
                        self.__ping_to_target(config.PING_CMD, target)
                        return True
                    except:
                        pass

            for target in [hostname] + inet6_addr_list:
                if target:
                    try:
                        self.__ping_to_target(config.PING6_CMD, target)
                        return True
                    except:
                        pass

        except:
            # Assume hostmame or IP address is used.
            return True

        return False

    def __switch_handler(self, entry):
        """Handle switch request.
           In Slave.
           Execute from WebUI.
           Switch Slave->Master(equivalent).
        <entry type="switch">
            <clone />
        </entry>
        """
        entry = entry.get_clone()
        entry.set_result(Results.OK)

        if not self.__is_valid_master_config():
            entry.set_result(Results.INVALID_DATA)
            return entry

        self.__try_update_service_for_switch()

        if self.__is_hostname_or_ip_used():
            entry.set_result(Results.INTERNAL_ERROR)
            return entry

        try:
            CmdExec(['/usr/local/bin/setslavexml', '--switch']).execute()
            entry.set_result(Results.OK)
        except:
            entry.set_result(Results.INTERNAL_ERROR)

        return entry
    # endfor MasterSlave ###

    # for SystemSave ###
    BACKUP_JOB_NAME = 'SystemRestore'
    LANDISKINFO_DATA_FILE_PATH = '/mnt/usb%s/%s/data' % \
                                    (USB_NO_STR, BACKUP_JOB_NAME)
    SYSTEM_BACKUP_SIGNATURE_PATH = '/mnt/usb%s/%s/signature' % \
                                    (USB_NO_STR, BACKUP_JOB_NAME)
    BACKUP_JOB_ID = 's1'
    BACKUP_JOB_DEST = 'usb%s' % (USB_NO_STR,)
    BACKUP_JOB_XML_FMT = """<backups><backup><job full="%(full)s">
        <name>%(name)s</name>
        <id>%(id)s</id>
        <times>%(times)s</times>
        <trash enable="false" />
        <shutdown enable="false" />
        <all_local enable="true" />
        <sourcelist />
        <destination>
            <unc>%(dest)s</unc>
        </destination>
    </job></backup></backups>"""

    BACKUP_JOB_NEW_XML_FMT = """<backups><backup><job full="%(full)s">
        <name>%(name)s</name>
        <id>%(id)s</id>
        <times>%(times)s</times>
        <trash enable="false" />
        <shutdown enable="false" />
        <all_local enable="true" />
        <sourcelist>
            <source>
                <type>all_local</type>
            </source>
        </sourcelist>
        <destination>
            <type>local</type>
            <sharename>%(dest)s</sharename>
            <subdir/>
        </destination>
    </job></backup></backups>"""

    def __get_backups_xml(self, full=False):
        try:
            times = self.__load_config()[SYSTEM][TIMES]
        except:
            times = 0

        series, model, version = util.get_sysinfo(False)

        backup_job_xml_fmt = self.BACKUP_JOB_XML_FMT
        if float(version) >= BACKUP_NEW_XML_FW_VER:
            backup_job_xml_fmt = self.BACKUP_JOB_NEW_XML_FMT

        return backup_job_xml_fmt % {
            'full': str(full).lower(),
            'times': times,
            'name': self.BACKUP_JOB_NAME,
            'id': self.BACKUP_JOB_ID,
            'dest': self.BACKUP_JOB_DEST,
        }

    def __load_backup_config_xml(self):
        try:
            landiskinfo_node = None
            with open(self.LANDISKINFO_DATA_FILE_PATH, 'r') as f:
                landiskinfo_node = XmlNode(self.__dec(f.read()))
                landiskinfo_node.delete_nodes(
                    'raidctrl/raidlist',
                    'raid/name[text()="sataraid1"]/..')
            return landiskinfo_node
        except:
            return None

    def __get_data_by_backup_and_restore(self, entry):
        data = dict()
        clone_node = entry.get()
        data['config'] = \
                clone_node.get_node('config').is_enabled() \
                if clone_node.get_node_list('config') \
                else False
        data['share'] = \
                clone_node.get_node('share').is_enabled() \
                if clone_node.get_node_list('share') \
                else False
        # NOTE restore時には不要
        data['full'] = \
                clone_node.get_node('share').is_set_to_true('full') \
                if clone_node.get_node_list('share') \
                else False
        return data

    def __get_pass(self):
        from subprocess import call, Popen, PIPE
        sdx = os.readlink('/dev/usbd%s' % (self.USB_NO_STR,))
        device = os.readlink('/sys/block/%s/device' % sdx)
        pass_ = ""
        for idX in ['idVendor', 'idProduct']:
            proc = Popen(
                ['cat',
                 '/sys/block/%(sdx)s/%(device)s/../../../../%(idX)s' % {
                     'idX': idX,
                     'device': device,
                     'sdx': sdx,
                 }],
                stdout=PIPE,
                stderr=PIPE)
            pass_ += proc.communicate()[0].strip()
        return pass_

    def __openssl(self, op, data):
        from subprocess import Popen, PIPE
        op = dict({'enc': '-e', 'dec': '-d'})[op]
        proc = Popen(
            ['/usr/bin/openssl', 'enc', '-aes-256-cbc', op,
             '-base64', '-pass', 'pass:%s' % self.__get_pass(), '-p'],
            stdin=PIPE,
            stdout=PIPE,
            stderr=PIPE,
        )
        (stdout, stderr) = proc.communicate(input=data)
        _data = ""
        for l in stdout.splitlines()[3:]:
            _data += "%s\n" % l
        return _data

    def __enc(self, data):
        from subprocess import call, Popen, PIPE
        proc = Popen(
            ['/usr/bin/openssl', 'enc', '-aes-256-cbc', '-e',
             '-base64', '-pass', 'pass:%s' % self.__get_pass(), '-p'],
            stdin=PIPE,
            stdout=PIPE,
            stderr=PIPE,
        )
        (stdout, stderr) = proc.communicate(input=data)
        encdata = ""
        for l in stdout.splitlines()[3:]:
            encdata += "%s\n" % l
        return encdata

    def __dec(self, encdata):
        from subprocess import call, Popen, PIPE
        proc = Popen(
            ['/usr/bin/openssl', 'enc', '-aes-256-cbc', '-d',
             '-base64', '-pass', 'pass:%s' % self.__get_pass(), '-p'],
            stdin=PIPE,
            stdout=PIPE,
            stderr=PIPE,
        )
        (stdout, stderr) = proc.communicate(input=encdata)
        data = ""
        for l in stdout.splitlines()[3:]:
            data += "%s\n" % l
        return data

    def __exists_usb(self):
        try:
            return os.path.exists('/dev/usbd%s' % (self.USB_NO_STR,))
        except:
            return False

    def __usb_admin_only(self):
        # FIXME イベントハンドラ
        try:
            entry = self.send_entry_with_file(
                    request.Entry(ReqTypes.LOAD),
                    pal_info.PAL_CONFFILE_PATH,
                    config.SHARES_CONF_FILE)
            if entry.get_result() != Results.OK:
                return False
            share_node = entry.get().get_node('sharelist/share[name="%s"]' % self.BACKUP_JOB_DEST)
            dp_node = share_node.get_node('permission/detailpermission')
            dp_node.enable(true=True)

            def __delete(parent, target):
                if target:
                    if parent.get_node_list(target[0]):
                        pass
                    else:
                        parent.append_child(XmlNode("<%s />" % target[0]))
                    __delete(parent.get_node(target[0]), target[1:])
                else:
                    parent.delete_children('*')

            for target in [
                    ('readonlymember', 'userlist'),
                    ('readonlymember', 'grouplist'),
                    ('readwritemember', 'userlist'),
                    ('readwritemember', 'grouplist'),
                    ]:
                __delete(dp_node, target)

            from fal_info import FAL_SHARES_PATH
            shares_node = XmlNode("<shares />")
            self.__delete_userinfo_for_share(share_node)
            shares_node.append_child(share_node)
            if self.send_entry(
                    request.Entry(ReqTypes.MODIFY, shares_node),
                    FAL_SHARES_PATH).get_result() == Results.OK:
                return True
            else:
                return False
        except:
            return False

    def __try_disable_dropbox(self, landiskinfo_node):
        """Try disable dropbox."""

        for dropbox_node in landiskinfo_node.get_node_list(DROPBOX_PATH):
            dropbox_node.enable(False)

    def __create_signature(self):
        """Create signature and return it."""

        return str(random.random())[2:]

    def __load_signature(self):
        """Return signature in configuration file.  If it does not exist,
           then create it.
        """

        clone_config = self.__load_config_file(self.CLONE_CONF_FILE).get()
        try:
            signature = clone_config.get_node(SIGNATURE).get_text()
        except:
            signature = self.__create_signature()
            clone_config.append_child(
                        XmlConfNode(SIGNATURE_XML_TMPL % signature))
            self.__save_config_file(self.CLONE_CONF_FILE, clone_config)

        return signature

    def __read_signature(self):
        """Read signature from a USB HDD."""

        try:
            with open(self.SYSTEM_BACKUP_SIGNATURE_PATH, 'r') as sign_file:
                return sign_file.read().strip()

        except:
            # Ignore error
            return ""

    def __write_signature(self):
        """Write signature to a USB HDD."""

        try:
            with open(self.SYSTEM_BACKUP_SIGNATURE_PATH, 'w') as sign_file:
                sign_file.write(self.__load_signature())

        except:
            # Ignore error
            pass

    def __backup_handler(self, entry):
        """Handle backup request.
           Execute from WebUI&cron.
           Backup "Setting data" and "Shared data".
        <entry type="backup">
            <clone>
                <config enable="false" />
                <share enable="true" />
            </clone>
        <entry>

        response result
        0(Results.OK)            : OK
        1                        : USB is not mounted
        2                        : USB is not share mode
        3                        : USB is not ext4
        4(Results.INTERNAL_ERROR): NG
        """
        entry = entry.get_clone()
        clone_node = entry.get()

        data = self.__get_data_by_backup_and_restore(entry)

        #if not self.__usb_is_backup_ok():
        #    self.__logging('ERR_HCLONE_SYSBACKUP_USB')
        #    entry.set_result("1")
        #    return entry
        if not self.__usb_is_mounted():
            self.__logging('ERR_HCLONE_SYSBACKUP_USB_DISCONNECT')
            entry.set_result("1")
            return entry
        if not self.__usb_is_share():
            self.__logging('ERR_HCLONE_SYSBACKUP_USB_NOT_SHARE')
            entry.set_result("2")
            return entry
        if not self.__usb_is_ext4():
            self.__logging('ERR_HCLONE_SYSBACKUP_USB_NOT_EXT4')
            entry.set_result("3")
            return entry

        system_backup_lock = FLock(config.LOCK_DIR, CLONE_SYSTEM_BACKUP_NAME,
                        exclusive=True, timeout=0)

        try:
            system_backup_lock.lock()
        except:
            system_backup_lock = None

        self.__usb_admin_only()

        success = True
        if data['config']:  # save config data
            from subprocess import call, Popen, PIPE
            proc = Popen(
                    ['/usr/local/bin/getinfoxml'],  # FIXME path
                    stdout=PIPE,
                    stderr=PIPE)
            (stdout, stderr) = proc.communicate()
            if proc.returncode != 0:
                success = False
            else:
                landiskinfo_node = XmlConfNode(stdout)
                for pxpath, cname in [
                        ('raidctrl/raidlist', 'raid'),
                        ('.', 'volctrl'),
                        ('servicectrl/servicelist', 'service/name[text()="narsus"]/..')]:
                    try:
                        landiskinfo_node.delete_nodes(pxpath, cname)
                    except:
                        if landiskinfo_node.get_node_list('/'.join([pxpath, cname])):
                            success = False

                self.__try_disable_dropbox(landiskinfo_node)

                if success:
                    try:
                        encdata = self.__enc(str(landiskinfo_node))
                        dirpath = os.path.dirname(
                                            self.LANDISKINFO_DATA_FILE_PATH)
                        if not os.path.exists(dirpath):
                            os.makedirs(dirpath)
                        with open(self.LANDISKINFO_DATA_FILE_PATH, 'w') as f:
                            f.write(encdata)
                    except:
                        success = False
        if data['share']:  # save share data
            if success:
                if self.send_entry(
                        request.Entry(
                                ReqTypes.EXECUTE,
                                self.__get_backups_xml(full=data['full'])),
                        pal_info.PAL_BACKUPS_PATH).get_result() not in [Results.OK, Results.QUEUE]:
                    success = False

            if success:
                self.__write_signature()
                self.__logging('MSG_HCLONE_SYSBACKUP_START')
            else:
                self.__logging('ERR_HCLONE_SYSBACKUP')

        if success:
            entry.set_result(Results.OK)
        else:
            entry.set_result(Results.INTERNAL_ERROR)  # FIXME code

        if system_backup_lock != None:
            system_backup_lock.unlock()

        #entry.set_result(Results.OK)
        return entry

    def __restore_handler(self, entry):
        """Handle backup request.
           Execute form WebUI.
           Restore "Setting data" and "Shared data".
        <entry type="restore">
            <clone>
                <config enable="false" />
                <share enable="true" generation="20140102030405" />
            </clone>
        <entry>
        """
        entry = entry.get_clone()
        clone_node = entry.get()

        data = self.__get_data_by_backup_and_restore(entry)
        data['generation'] = entry.get().get_value('share/@generation', None)

        #if not self.__exists_usb():
        if not self.__usb_is_restore_ok():
            self.__logging('ERR_HCLONE_SYSRESTORE_USB')
            entry.set_result("1")
            return entry

        success = True
        if data['config']:  # restore config data
#            landiskinfo_node = None
#            with open(self.LANDISKINFO_DATA_FILE_PATH, 'r') as f:
#                landiskinfo_node = XmlNode(self.__dec(f.read()))
#                landiskinfo_node.delete_nodes(
#                    'raidctrl/raidlist',
#                    'raid/name[text()="sataraid1"]/..')
            landiskinfo_node = self.__load_backup_config_xml()
            if landiskinfo_node:
                from subprocess import call, Popen, PIPE
                proc = Popen(
                    ['/usr/local/bin/setinfoxml'],  # FIXME path
                    stdin=PIPE)
                proc.communicate(str(landiskinfo_node))
                if proc.returncode != 0:
                    success = False
            else:
                success = False
        if success and data['share']:  # restore share data
            backups_node = XmlNode(self.__get_backups_xml())
            if data['generation']:
                backups_node.get_node('backup/job').append_child(XmlNode(
                        """<generation>
                            <id>%s</id>
                        </generation>""" % data['generation']))
            if self.send_entry(
                    request.Entry(ReqTypes.RESTORE, backups_node),
                    pal_info.PAL_BACKUPS_PATH).get_result() != Results.OK:
                success = False
        if success:
            entry.set_result(Results.OK)
        else:
            entry.set_result(Results.INTERNAL_ERROR)  # FIXME code

        if success:
            self.__logging('MSG_HCLONE_SYSRESTORE_START')

            str(landiskinfo_node) | CmdExec(
                    ['/usr/local/bin/setnetworkinfoxml'],  # FIXME path
                    wait=False)
        else:
            self.__logging('ERR_HCLONE_SYSRESTORE')
        #entry.set_result(Results.OK)
        return entry
    # endfor SystemSave ###

    def __stop_handler(self, entry):
        try:
            if self.send_entry(
                    request.Entry('stop', """<backups>
                        <backup>
                            <job>
                                <id>%s</id>
                            </job>
                        </backup>
                    </backups>""" % self.BACKUP_JOB_ID),
                    pal_info.PAL_BACKUPS_PATH).get_result() == Results.OK:
                entry.set_result(Results.OK)
            else:
                entry.set_result(Results.INTERNAL_ERROR)  # FIXME code
            return entry
        except:
            entry.set_result(Results.INTERNAL_ERROR)
            return entry

    def __init__(self, option=None):
        """Constructor."""

        falcmd.FalCmd.__init__(self, option=option)

        self.set_handler('query', self.__query_handler)
        self.set_handler('apply', self.__apply_handler)

        self.set_handler('save', self.__save_handler)
        self.set_handler('switch', self.__switch_handler)

        self.set_handler('backup', self.__backup_handler)
        self.set_handler('restore', self.__restore_handler)

        self.set_handler('stop', self.__stop_handler)


#
# Main
#
if __name__ == "__main__":
    FalClone(option=falcmd.FalCmdOption()).handle_request()
