#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""
Handler to backup configuration for events which are issued when configuration
is changed.
"""

from iobase.fileut import (FLock,)
from fal_info import (FAL_CLONE_PATH, Results, ReqTypes)
from fal_msginfo import (Msg,)
from ionas import (request,)
from libevh import (evhcmd, util)
from evh_config import (LOCK_DIR,)
from pkgaddon.evh_hclone_config import (CLONE_SYSTEM_BACKUP_NAME,)


EVENTINFO_NAME = "eventinfo"
PREVIOUS_ATTR = "previous"

CLONE_NAME = "clone"
SIGNATURE_PATH = "signature"
EXISTS_ATTR = "exists"

CLONE_QUERY_XML = "<clone/>"
CLONE_SAVE_CONFIG_XML = """<clone><config enable="true"/></clone>"""


class EvhCloneBackupConf(evhcmd.EvhCmd):
    """Event handler for the Clone packge."""

    def __is_there_backup_config(self):
        """Check if there is the backuped config."""

        try:
            signature_node = self.send_entry(
                            request.Entry(ReqTypes.QUERY, CLONE_QUERY_XML),
                            FAL_CLONE_PATH).get().get_node(SIGNATURE_PATH)
            if signature_node.is_set_to_true(EXISTS_ATTR):
                return True
        except:
            pass

        return False

    def __backup_config(self, eventinfo):
        """Backup the landiskinfo XML data to the USB HDD."""

        # There is typo "BUCKUP" in product_conf.reqinfo.
        response_entry = self.send_entry(
                        request.Entry(ReqTypes.BUCKUP, CLONE_SAVE_CONFIG_XML),
                        FAL_CLONE_PATH)
        if response_entry.get_result() != Results.OK:
            raise util.RequestError(response_entry.get_result())

    def __event_handler(self, entry):
        """Handle "event" request."""

        try:
            eventinfo = entry.get().get_node(EVENTINFO_NAME)
            if eventinfo.is_set_to_true(PREVIOUS_ATTR):
                entry.set_result(Results.OK)
                return entry

            evh_lock = FLock(LOCK_DIR, CLONE_SYSTEM_BACKUP_NAME,
                        exclusive=True, timeout=0)
            try:
                evh_lock.lock()
            except:
                # Event is for change share configuration by fal_clone. 
                # Ignore it.
                entry.set_result(Results.OK)
                return entry

            try:
                if self.__is_there_backup_config():
                    self.__backup_config(eventinfo)
                    util.save_event_log(Msg.MSG_HCLONE_SAVE_CONFIG)
            except:
                self.exception("Backup config")
                util.save_event_log(Msg.ERR_HCLONE_SAVE_CONFIG)
                # Go through to set a result to "OK".
            finally:
                evh_lock.unlock()
        except:
            # Non target events
            pass

        entry.set_result(Results.OK)
        return entry

    def __init__(self, option=None):
        """Constructor."""

        evhcmd.EvhCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.EVENT, self.__event_handler)


#
# Main
#

if __name__ == "__main__":
    EvhCloneBackupConf(option=evhcmd.EvhCmdOption()).handle_request()
